package net.sf.tomp.xml.include;

import net.sf.tomp.xml.type.Type;
import net.sf.tomp.xml.type.TypeDatabase;
import net.sf.tomp.xml.type.TypeDetectorFilter;
import net.sf.tomp.xml.type.TypeTransformation;
import net.sf.tomp.xml.type.TypeTransformationException;
import net.sf.tomp.xml.type.Variant;
import net.sf.tomp.xml.type.VariantImpl;
import net.sf.tomp.xtcl.filter.DynamicFilterImpl;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xml.sax.SAXException;
import org.xml.sax.XMLFilter;
import org.xml.sax.XMLReader;

/**
 * DOCUMENT ME!
 * 
 * @author tomp
 */
public class AdaptiveFilterImpl extends DynamicFilterImpl {
    ////////////////////////////////////////////////////////////////////
    // Constructors.
    ////////////////////////////////////////////////////////////////////
    private Log log = LogFactory.getLog(AdaptiveFilterImpl.class);

    /** DOCUMENT ME! */
    protected TypeDetectorFilter typeDetectorFilter;

    /** DOCUMENT ME! */
    protected Type targetType;

    /** DOCUMENT ME! */
    protected Variant variant = Variant.NO_VARIANT;

    /**
     * Creates a new AdaptiveFilterImpl object.
     */
    public AdaptiveFilterImpl() {
        super();
        typeDetectorFilter = new TypeDetectorFilter();
        parent = typeDetectorFilter;
    }

    /**
     * Construct an XML filter with the specified parent.
     * 
     * @param p DOCUMENT ME!
     * @see #setParent
     * @see #getParent
     */
    public AdaptiveFilterImpl(XMLReader p) {
        super();
        typeDetectorFilter = new TypeDetectorFilter(p);
        parent = typeDetectorFilter;
    }

    /**
     * DOCUMENT ME!
     * 
     * @return DOCUMENT ME!
     */
    public TypeDatabase getTypeDatabase() {
        return typeDetectorFilter.getTypeDatabase();
    }

    /**
     * DOCUMENT ME!
     * 
     * @param td DOCUMENT ME!
     */
    public void setTypeDatabase(TypeDatabase td) {
        typeDetectorFilter.setTypeDatabase(td);
    }

    /**
     * DOCUMENT ME!
     * 
     * @return DOCUMENT ME!
     */
    public Type getSourceType() {
        return typeDetectorFilter.getType();
    }

    /**
     * DOCUMENT ME!
     * 
     * @return DOCUMENT ME!
     */
    public Type getSourceTypeForCurrentContext() {
        return typeDetectorFilter.getTypeForCurrentContext();
    }

    /**
     * DOCUMENT ME!
     * 
     * @return DOCUMENT ME!
     */
    public Type getTargetType() {
        return targetType;
    }

    /**
     * DOCUMENT ME!
     * 
     * @param t DOCUMENT ME!
     */
    public void setTargetType(Type t) {
        targetType = t;
    }

    /**
     * DOCUMENT ME!
     * 
     * @return DOCUMENT ME!
     */
    public Variant getVariant() {
        return variant;
    }

    /**
     * DOCUMENT ME!
     * 
     * @param t DOCUMENT ME!
     */
    public void setVariant(Variant t) {
        variant = t;
    }

    /**
     * DOCUMENT ME!
     * 
     * @param identification DOCUMENT ME!
     */
    public void setTargetType(String identification) {
        Type targetType = getTypeDatabase().getType(identification);

        setTargetType(targetType);
    }

    // begin of the XTFilter impl. (overriden)
    public void setParameter(String k, Object v) {
        //System.out.println("Setting param "+k+"="+v);
        if ((k != null) && k.startsWith("target-")) {
            String tt = k.substring("target-".length()) + ":" + (String) v;

            setTargetType(tt);

            //System.out.println("*** Target type was set
            // to="+getTargetType());
        } else if ((k != null) && k.equals("variant")) {
            setVariant(new VariantImpl((String) v));
        } else {
            super.setParameter(k, v);
        }
    }

    // end of the XTFilter impl.
    // begin of XMLFilter impl.
    public void setParent(XMLReader parent) {
        super.setParent(typeDetectorFilter);

        //System.out.println(" $$$ "+this+" setParent to
        // "+typeDetectorFilter+"... ");
        if (typeDetectorFilter != null) {
            //  System.out.println(" set to
            // typeDetectorFilter="+typeDetectorFilter);
            //  System.out.println(" own parent is="+getParent());
            typeDetectorFilter.setParent(parent);

            //        } else {
            //            System.out.println(" $$$ AdaptiveFilter.setParent to "+parent+"
            // but typeDetector is null in"+this);
        }
    }

    // end of XMLFilter impl.

    /**
     * Filter method startDocument finds and activate the filter for the
     * transformation
     * 
     * @throws SAXException DOCUMENT ME!
     */
    public void startDocument() throws SAXException {
        //System.out.println(""+this+".startDocument");
        TypeDatabase tdb = getTypeDatabase();
        Type srcType = getSourceType();
        Type targetType = getTargetType();

        log.info("\nDETECTED src" + srcType + "\nDETECTED tgt" + targetType);

        if (variant != null) {
            log.info("VARIANT  " + variant + " tdb=" + tdb);
        }

        if (targetType != null) {
            TypeTransformation tt = tdb.getTransformation(srcType, targetType,
                    variant);

            log.info("TRANSFORM " + tt);

            if (tt != null) {
                try {
                    XMLFilter tf = tt.newFilter();

                    setFilter(tf);
                } catch (TypeTransformationException tte) {
                    throw new SAXException(tte);
                }

                //System.out.println("&&&&&& filter="+tf);

                /*
                 * if(tf instanceof TransformerFilterImpl) {
                 * ((TransformerFilterImpl)tf).getResult().setSystemId("file:something");
                 * System.out.println("&&&&&& filter="+tf+" has
                 * Result="+((TransformerFilterImpl)tf).getResult());
                 * System.out.println("&&&&&& filter="+tf+" has
                 * Result.getSystemId()="+((TransformerFilterImpl)tf).getResult().getSystemId()); }
                 */

                //System.out.println("*** filter has been set");
            }
        }

        super.startDocument();
    }
}

/*
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at http://www.mozilla.org/MPL/
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License. The
 * Original Code is: all this file. The Initial Developer of the Original Code
 * is: Tomas Pitner, Masaryk University in Brno, Czech Republic. Contributor(s):
 */
