/*
 * AdaptiveXIncluder.java Created on 25. duben 2004, 0:10
 */
package net.sf.tomp.xml.include;

import net.sf.tomp.xml.type.TypeDatabaseImpl;
import net.sf.tomp.xtcl.Context;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.PosixParser;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Result;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.sax.SAXSource;
import javax.xml.transform.sax.SAXTransformerFactory;
import javax.xml.transform.sax.TransformerHandler;
import javax.xml.transform.stream.StreamResult;

import java.io.IOException;

/**
 * @author tomp
 */
public class AdaptiveXIncluder {
    private boolean verbose;

    private String tdFileName;

    private String inputFileName;

    private String outputFileName;

    private TypeDatabaseImpl tdb;

    private Context context;

    private AdaptiveXIncludeFilter axif;

    private SAXTransformerFactory stf;

    /**
     * Creates a new instance of AdaptiveXIncluder
     * 
     * @throws TransformerConfigurationException
     * @throws ParserConfigurationException
     * @throws IOException
     * @throws ClassNotFoundException
     * @throws IllegalAccessException
     * @throws InstantiationException
     */
    public AdaptiveXIncluder() throws TransformerConfigurationException,
            ParserConfigurationException, IOException, InstantiationException,
            IllegalAccessException, ClassNotFoundException {
        context = new Context();
    }

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) throws Exception {
        String tdFileName = "type-database.xml";

        OptionBuilder.hasArgs(1);
        OptionBuilder.withLongOpt("type-database");
        OptionBuilder
                .withDescription("use the specified file as Type Database");
        OptionBuilder.withArgName("file");

        Option oTdb = OptionBuilder.create('d');

        OptionBuilder.withLongOpt("verbose");
        OptionBuilder
                .withDescription("provide more detailed info on processing");

        Option oVerb = OptionBuilder.create('v');

        Options opts = new Options();

        opts.addOption(oTdb);
        opts.addOption(oVerb);

        CommandLine cl = new PosixParser().parse(opts, args);

        AdaptiveXIncluder axi = new AdaptiveXIncluder();

        try {
            axi.setInputFileName(cl.getArgs()[0]);
        } catch (ArrayIndexOutOfBoundsException aiob) {
            new HelpFormatter()
                    .printHelp(
                            "Adaptive XML Includer, the axi:include references in infile are processes using the Type Database",
                            opts);

            return;
        }

        if (cl.getArgs().length > 1) {
            axi.setOutputFileName(cl.getArgs()[1]);
        }

        axi.setVerbose(cl.hasOption('v'));
        axi.setTypeDatabaseFileName(cl.hasOption('d') ? cl.getOptionValue('d')
                : tdFileName);
        axi.include();
    }

    public void include() throws IOException, TransformerException,
            TransformerConfigurationException, ParserConfigurationException,
            SAXException {
        tdb = new TypeDatabaseImpl();
        tdb.setXTFilterFactory(context.getXTFilterFactory());

        // load Type Database
        XMLReader reader = getXMLReader();

        reader.setContentHandler(tdb);
        reader.parse(tdFileName);

        axif = new AdaptiveXIncludeFilter(getXMLReader());
        axif.setTypeDatabase(tdb);

        TransformerFactory tf = context.getTransformerFactory();

        if (!tf.getFeature(SAXSource.FEATURE)) {
            throw new TransformerException("Not a SAXTransformerFactory");
        }

        stf = (SAXTransformerFactory) tf;

        Result result = (outputFileName == null) ? new StreamResult(System.out)
                : new StreamResult(context.getFile(outputFileName));

        TransformerHandler th = stf.newTransformerHandler();

        axif.setContentHandler(th);
        axif.setProperty("http://xml.org/sax/properties/lexical-handler", th);
        th.setResult(result);

        if (verbose) {
            System.out.println("NOW PARSING source=" + inputFileName);
        }

        axif.parse(context.getFile(inputFileName).toString());
    }

    protected XMLReader getXMLReader() throws SAXException {
        XMLReader parser = null;

        try {
            parser = XMLReaderFactory.createXMLReader();
        } catch (SAXException e) {
            parser = XMLReaderFactory.createXMLReader(

            //"org.apache.xerces.parsers.SAXParser"
                    "org.apache.crimson.parser.XMLReaderImpl");
        }

        return parser;
    }

    public void setInputFileName(String ifn) {
        inputFileName = ifn;
    }

    public void setOutputFileName(String ofn) {
        outputFileName = ofn;
    }

    public void setVerbose(boolean v) {
        verbose = v;
    }

    public void setTypeDatabaseFileName(String tdfn) {
        tdFileName = tdfn;
    }
}