package net.sf.tomp.xml.type;

import net.sf.tomp.xtcl.filter.XTFilterFactory;
import net.sf.tomp.xtcl.result.HasResult;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.XMLFilter;
import org.xml.sax.helpers.DefaultHandler;

import javax.xml.transform.Result;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.sax.SAXResult;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;

/**
 * DOCUMENT ME!
 * 
 * @author $author$
 * @version $Revision$
 */
public class TypeDatabaseImpl extends DefaultHandler /* XTFilterImpl */
implements TypeDatabase, TypeTransformationFactory, HasResult {
    //protected Set allTypes = new HashSet();
    private Log log = LogFactory.getLog(TypeDatabaseImpl.class);

    /** DOCUMENT ME! */
    protected XTFilterFactory xff;

    /**
     * parent type database. If it is set, then all queries not answered here
     * are forwarded to parent TD
     */
    protected TypeDatabase parentDatabase;

    /** DOCUMENT ME! */
    protected Type currentType;

    /** DOCUMENT ME! */
    protected Type currentSrcType;

    /** DOCUMENT ME! */
    protected Type currentTargetType;

    /** DOCUMENT ME! */
    protected String currentSource;

    /** DOCUMENT ME! */
    protected String currentTarget;

    /** DOCUMENT ME! */
    protected TypeTransformation currentTransformation;

    /** DOCUMENT ME! */
    protected MultiFilterTypeTransformation currentMultiTransformation;

    /** DOCUMENT ME! */
    protected Variant currentVariant;

    /** DOCUMENT ME! */
    protected Map identification2type = new HashMap();

    /** DOCUMENT ME! */
    protected Map root2type = new HashMap();

    /** DOCUMENT ME! */
    protected Map doctype2type = new HashMap();

    // array of all types in the TypeDB
    //protected Type[] allTypes;
    //protected Map srcTransformations = new HashMap();
    //protected Map transformations = new HashMap();

    /** DOCUMENT ME! */
    protected Transformations transformations = new Transformations();

    /**
     * DOCUMENT ME!
     * 
     * @param x DOCUMENT ME!
     */
    public void setXTFilterFactory(XTFilterFactory x) {
        xff = x;
    }

    /**
     * DOCUMENT ME!
     * 
     * @return DOCUMENT ME!
     */
    public TypeDatabase getParent() {
        return parentDatabase;
    }

    /**
     * DOCUMENT ME!
     * 
     * @param td DOCUMENT ME!
     */
    public void setParent(TypeDatabase td) {
        parentDatabase = td;
    }

    /**
     * DOCUMENT ME!
     * 
     * @return DOCUMENT ME!
     */
    public Result getResult() {
        return new SAXResult(this);
    }

    /**
     * DOCUMENT ME!
     * 
     * @return DOCUMENT ME!
     */
    public int getTypesCount() {
        return identification2type.entrySet().size();
    }

    /**
     * DOCUMENT ME!
     * 
     * @return DOCUMENT ME!
     */
    public Transformations getTransformations() {
        return transformations;
    }

    /**
     * implementation of transformation info storage
     * 
     * @param source DOCUMENT ME!
     * @param target DOCUMENT ME!
     * @param variant DOCUMENT ME!
     * @param tt DOCUMENT ME!
     */

    /*
     * public void addTransformation(Type source, Type target,
     * TypeTransformation tt) { addTransformation(new TransformationKey(source,
     * target), tt); }
     */
    public void addTransformation(Type source, Type target, Variant variant,
            TypeTransformation tt) {
        /*
         * Map transformations = (Map)srcTransformations.get(source);
         * if(transformations == null) { transformations = new HashMap();
         * srcTransformations.put(source, transformations); }
         * transformations.put(target, tt);
         */
        TransformationsFromType tft = transformations
                .getTransformationsFrom(source);

        if (tft == null) {
            tft = new TransformationsFromType(source);
            transformations.putTransformationsFrom(tft);
        }

        tft.putTransformationTo(target, variant, tt);
    }

    /**
     * DOCUMENT ME!
     * 
     * @param source DOCUMENT ME!
     * @param target DOCUMENT ME!
     * @param variant DOCUMENT ME!
     * @return DOCUMENT ME!
     */
    public TypeTransformation getTransformation(Type source, Type target,
            Variant variant) {
        if (source == target) {
            return null;
        }

        /*
         * Map transformations = (Map)srcTransformations.get(source);
         * if(transformations != null) { Object o = transformations.get(target);
         * if(o == null && parentDatabase != null) return
         * parentDatabase.getTransformation(source, target, variant); else
         * return (TypeTransformation)o; } else if(parentDatabase != null) {
         * return parentDatabase.getTransformation(source, target, variant); }
         * return null;
         */
        TransformationsFromType tft = transformations
                .getTransformationsFrom(source);

        if (tft == null) {
            if (parentDatabase == null) {
                return null;
            } else {
                return parentDatabase
                        .getTransformation(source, target, variant);
            }
        }

        TypeTransformation tt = tft.getTransformationTo(target, variant);

        if (tt == null) {
            log.debug("+++ SEARCHING transformationSequence to=" + target
                    + " variant=" + variant);

            List transformationSequence = getTransformationSequence(source,
                    target, variant);

            log.debug("+++ FOUND transformationSequence="
                    + transformationSequence);

            if ((transformationSequence == null)
                    || (transformationSequence.size() == 0)) {
                return null;
            }

            TypeTransformation mt = new MultiFilterTypeTransformation(
                    transformationSequence);

            // caching!
            tft.putTransformationTo(target, variant, mt);

            return mt;
        }

        return tft.getTransformationTo(target, variant);
    }

    /*
     * protected void addTransformation(TransformationKey k, TypeTransformation
     * tt) { transformations.put(k, tt); } public TypeTransformation
     * getTransformation(Type source, Type target) { return
     * getTransformation(new TransformationKey(source, target)); } public
     * TypeTransformation getTransformation(Type source, Type target, Variant
     * variant) { return getTransformation(new TransformationKey(source, target,
     * variant)); }
     */
    /*
     * public TypeTransformation getTransformation(TransformationKey k) {
     * TypeTransformation tt = (TypeTransformation)transformations.get(k); if
     * (tt == null && parentDatabase != null && k != null) return
     * parentDatabase.getTransformation(k.source, k.target, k.variant); return
     * tt; }
     */

    /**
     * implementation of type info storage
     * 
     * @param ns DOCUMENT ME!
     * @param localName DOCUMENT ME!
     * @return DOCUMENT ME!
     */
    public Type getTypeForRootElement(String ns, String localName) {
        Object t = root2type.get("{" + ns + "}" + localName);

        if (t == null) {
            t = root2type.get("{" + ns + "}*");
        }

        if (t == null) {
            t = root2type.get("{*}" + localName);
        }

        return (Type) t;
    }

    /**
     * DOCUMENT ME!
     * 
     * @param name DOCUMENT ME!
     * @param dtPublic DOCUMENT ME!
     * @param dtSystem DOCUMENT ME!
     * @return DOCUMENT ME!
     */
    public Type getTypeForDoctype(String name, String dtPublic, String dtSystem) {
        return (Type) doctype2type.get(dtPublic);
    }

    /**
     * DOCUMENT ME!
     * 
     * @param id DOCUMENT ME!
     * @return DOCUMENT ME!
     */
    public Type getType(String id) {
        return (Type) identification2type.get(id);
    }

    /**
     * DOCUMENT ME!
     * 
     * @param fn DOCUMENT ME!
     * @return DOCUMENT ME!
     * @throws ClassNotFoundException DOCUMENT ME!
     * @throws InstantiationException DOCUMENT ME!
     * @throws IllegalAccessException DOCUMENT ME!
     */
    public TypeTransformation newTransformationForFilterClassName(String fn)
            throws ClassNotFoundException, InstantiationException,
            IllegalAccessException {
        return new FilterTypeTransformation(fn);
    }

    /**
     * DOCUMENT ME!
     * 
     * @param ssi DOCUMENT ME!
     * @return DOCUMENT ME!
     * @throws TransformerConfigurationException DOCUMENT ME!
     */
    public TypeTransformation newTransformationForStyleSystemId(String ssi)
            throws TransformerConfigurationException {
        //System.out.println("&&&&&&&&& returning transformation for
        // stylehref="+ssi);
        return new StyleTypeTransformation(xff, ssi);
    }

    /**
     * DOCUMENT ME!
     * 
     * @param ssi DOCUMENT ME!
     * @return DOCUMENT ME!
     * @throws TransformerConfigurationException DOCUMENT ME!
     */
    public TypeTransformation newTransformationForSTXStyleSystemId(String ssi)
            throws TransformerConfigurationException {
        //System.out.println("&&&&&&&&& returning transformation for
        // stylehref="+ssi);
        return new StyleTypeTransformation(xff, ssi);
    }

    // end of TypeTransformationFactory implementation

    /**
     * Filter a start document event.
     * 
     * @param uri DOCUMENT ME!
     * @param localName DOCUMENT ME!
     * @param qName DOCUMENT ME!
     * @param atts DOCUMENT ME!
     * @exception SAXException The client may throw an exception during
     *                processing.
     */

    /*
     * public void startDocument () throws SAXException {
     * //System.out.println(""+this+".startDocument"); super.startDocument(); }
     */

    /**
     * Filter an end document event.
     * 
     * @param uri DOCUMENT ME!
     * @param localName DOCUMENT ME!
     * @param qName DOCUMENT ME!
     * @param atts DOCUMENT ME!
     * @exception SAXException The client may throw an exception during
     *                processing.
     */

    /*
     * public void endDocument () throws SAXException {
     * //System.out.println(""+this+".endDocument"); Type source =
     * getTypeForPublicDoctype("Type1"); Type target =
     * getTypeForPublicDoctype("Type4");
     * System.out.println("transformationSequence from "+source+" to "+target+"
     * is " +getTransformationSequence(source, target, Variant.NO_VARIANT));
     * super.endDocument(); }
     */

    /**
     * Filter a start element event.
     * 
     * @param uri The element's Namespace URI, or the empty string.
     * @param localName The element's local name, or the empty string.
     * @param qName The element's qualified (prefixed) name, or the empty
     *            string.
     * @param atts The element's attributes.
     * @exception SAXException The client may throw an exception during
     *                processing.
     */
    public void startElement(String uri, String localName, String qName,
            Attributes atts) throws SAXException {
        XMLFilter filter = null;

        if ("type".equals(qName)) {
            for (int i = 0; i < atts.getLength(); i++) {
                String aName = atts.getQName(i);

                if (aName.startsWith("by-")) {
                    String typeType = aName.substring("by-".length());
                    String identification = atts.getValue(i);

                    currentType = new TypeImpl(typeType, identification);
                    identification2type.put(typeType + ":" + identification,
                            currentType);

                    break;
                }
            }
        } else if ("subtype-of".equals(qName) || "parent-is".equals(qName)) {
            String typeIdentification = null;

            for (int i = 0; i < atts.getLength(); i++) {
                String aName = atts.getQName(i);

                if (aName.startsWith("by-")) {
                    String typeType = aName.substring("by-".length());
                    String identification = atts.getValue(i);

                    typeIdentification = typeType + ":" + identification;

                    break;
                }
            }

            if (typeIdentification != null) {
                Type parentType = getType(typeIdentification);

                if ((currentType != null) && (parentType != null)
                        && (currentType != parentType)) {
                    currentVariant = new VariantImpl(atts.getValue("variant"));

                    currentType.addParent(parentType);
                    log.debug(">>> add IDENTITY from  currentType="
                            + currentType + " to " + parentType);

                    addTransformation(currentType, parentType, currentVariant,
                            TypeTransformation.IDENTITY);
                } else {
                    System.err.println("Parent type '" + typeIdentification
                            + "' does not exists yet or is same as current: "
                            + currentType);
                }
            }
        } else if ("part-of".equals(qName)) {
            String typeIdentification = null;

            for (int i = 0; i < atts.getLength(); i++) {
                String aName = atts.getQName(i);

                if (aName.startsWith("by-")) {
                    String typeType = aName.substring("by-".length());
                    String identification = atts.getValue(i);

                    typeIdentification = typeType + ":" + identification;

                    break;
                }
            }

            String contextIdentification = atts.getValue("context");

            if ((typeIdentification != null) && (contextIdentification != null)) {
                TypeContextImpl context = new TypeContextImpl(
                        contextIdentification);
                Type parentType = getType(typeIdentification);

                if ((currentType != null) && (parentType != null)
                        && (context != null) && (currentType != parentType)) {
                    log.debug(">>> parentType= " + parentType
                            + " storing context=" + context + " currentType="
                            + currentType);
                    parentType.addTypeForContext(context, currentType);
                } else {
                    System.err.println("Whole-doc type '" + typeIdentification
                            + "' does not exists yet or is same as current: "
                            + currentType + "for context " + context);
                }
            }
        } else if ("root".equals(qName)) {
            String name = atts.getValue("name");

            if (name == null) {
                name = "*";
            }

            String ns = atts.getValue("ns");

            if (ns == null) {
                ns = "*";
            }

            if (currentType != null) {
                root2type.put("{" + ns + "}" + name, currentType);
            }
        } else if ("doctype".equals(qName)) {
            String name = atts.getValue("name");

            if (name == null) {
                name = "";
            }

            String pub = atts.getValue("public");

            if (pub == null) {
                pub = "";
            }

            String syst = atts.getValue("system");

            if (syst == null) {
                syst = "";
            }

            if (currentType != null) {
                // FIXME - works for PUBLIC DOCTYPE only!
                doctype2type.put(pub, currentType);
            }
        } else if ("transformation".equals(qName)) {
            for (int i = 0; i < atts.getLength(); i++) {
                String aName = atts.getQName(i);

                if (aName.startsWith("source-")) {
                    String typeType = aName.substring("source-".length());
                    String identification = atts.getValue(i);
                    String typeIdentification = typeType + ":" + identification;

                    currentSrcType = getType(typeIdentification);
                } else if (aName.startsWith("target-")) {
                    String typeType = aName.substring("target-".length());
                    String identification = atts.getValue(i);
                    String typeIdentification = typeType + ":" + identification;

                    currentTargetType = getType(typeIdentification);
                }
            }

            currentVariant = new VariantImpl(atts.getValue("variant"));

            currentTransformation = null;
            currentMultiTransformation = null;
        } else if ("filter".equals(qName)) {
            String className = atts.getValue("class");

            //System.out.println("making transformation for filter class
            // "+className);
            try {
                if (currentTransformation != null) {
                    // already was a transformation (i.e. it is multiple)
                    if (currentMultiTransformation == null) {
                        // this is the second
                        currentMultiTransformation = new MultiFilterTypeTransformation();
                    }

                    // adding last to multiple
                    currentMultiTransformation
                            .addTransformation(currentTransformation);
                }

                currentTransformation = newTransformationForFilterClassName(className);
            } catch (Exception cnfe) {
                cnfe.printStackTrace();
            }
        } else if ("style".equals(qName)) {
            // FIXME - to be completed
            String styleHref = atts.getValue("href");
            String styleType = atts.getValue("type");

            try {
                if (currentTransformation != null) {
                    // already was a transformation (i.e. it is multiple)
                    if (currentMultiTransformation == null) {
                        // this is the second
                        currentMultiTransformation = new MultiFilterTypeTransformation();
                    }

                    // adding last to multiple
                    currentMultiTransformation
                            .addTransformation(currentTransformation);
                }

                if ((styleType == null) || styleType.startsWith("xsl")) {
                    currentTransformation = newTransformationForStyleSystemId(styleHref);
                } else if (styleType.startsWith("stx")
                        || styleType.startsWith("joost")) {
                    currentTransformation = newTransformationForSTXStyleSystemId(styleHref);
                }

                //throw new IllegalArgumentException("STX styles are not
                // supported yet.");
            } catch (TransformerConfigurationException tce) {
                // FIXME - produces identity filter
                tce.printStackTrace();

                //throw new
                // IllegalArgumentException("TransformerConfigurationException:
                // currentTransformation = null");
            }
        } else if ("param".equals(qName)) {
            if (currentTransformation != null) {
                String pName = atts.getValue("name");

                // FIXME: value can be specified as attribute only
                String pValue = atts.getValue("value");

                currentTransformation.setParameter(pName, pValue);
            }
        }

        super.startElement(uri, localName, qName, atts);
    }

    /**
     * Filter an end element event.
     * 
     * @param uri The element's Namespace URI, or the empty string.
     * @param localName The element's local name, or the empty string.
     * @param qName The element's qualified (prefixed) name, or the empty
     *            string.
     * @exception SAXException The client may throw an exception during
     *                processing.
     */
    public void endElement(String uri, String localName, String qName)
            throws SAXException {
        if ("type".equals(qName)) {
            currentType = null;
        } else if ("transformation".equals(qName)) {
            if (currentMultiTransformation == null) {
                addTransformation(currentSrcType, currentTargetType,
                        currentVariant, currentTransformation);
            } else {
                currentMultiTransformation
                        .addTransformation(currentTransformation);
                addTransformation(currentSrcType, currentTargetType,
                        currentVariant, currentMultiTransformation);
            }
        } else if ("filter".equals(qName)) {
        } else if ("style".equals(qName)) {
        } else if ("param".equals(qName)) {
        }

        super.endElement(uri, localName, qName);
    }

    /**
     * Filter a character data event.
     * 
     * @param ch An array of characters.
     * @param start The starting position in the array.
     * @param length The number of characters to use from the array.
     * @exception SAXException The client may throw an exception during
     *                processing.
     */
    public void characters(char[] ch, int start, int length)
            throws SAXException {
        super.characters(ch, start, length);
    }

    /**
     * DOCUMENT ME!
     * 
     * @param source DOCUMENT ME!
     * @param target DOCUMENT ME!
     * @param variant DOCUMENT ME!
     * @return DOCUMENT ME!
     */
    public List getTransformationSequence(final Type source, final Type target,
            Variant variant) {
        // FIXME: no variant
        List sequence = new ArrayList();

        //DynamicMultiFilterImpl dmf = new DynamicMultiFilterImpl();
        //Set pivots = new HashSet();
        SortedSet typeNodes = new TreeSet();
        Map types2TypeNodes = new HashMap();

        // init pivot
        TypeNode pivot = new TypeNode(source);

        pivot.costsFromSource = 0;
        pivot.wasPivot = true;
        types2TypeNodes.put(source, pivot);

        do {
            // add new pivot
            //System.out.println("&&& new pivot is "+pivot);
            //pivots.add(pivot);
            // get all transf. from pivot type
            TransformationsFromType transFromPivot = transformations
                    .getTransformationsFrom(pivot.typ);

            //System.out.println("SEEKING trans from "+source+" to "+target);
            log.info("\n\nEXAMINING transf from pivot=" + pivot.typ
                    + " costs to pivot=" + pivot.costsFromSource
                    + " ... transFromPivot=" + transFromPivot);

            // FIXME: if transFromPivot == null???
            if (transFromPivot == null) {
                ;
            }

            //System.out.println(" NO TRANSF ");
            else {
                for (Iterator i = transFromPivot.typesIterator(); i.hasNext();) {
                    // target type
                    Type tg = (Type) i.next();

                    // transformation to target type
                    TypeNode tnOrig = (TypeNode) types2TypeNodes.get(tg);

                    TypeTransformation ts = transFromPivot.getTransformationTo(
                            tg, variant);

                    // if no transformation, i.e. not THE variant... skip to
                    // next
                    if (ts == null) {
                        continue;
                    }

                    TypeNode tn = (tnOrig == null) ? new TypeNode(tg) : tnOrig;

                    log.info("   TRANSF TO " + tg + " is " + ts);

                    // doing only those that haven't been pivots
                    if (!tn.wasPivot) {
                        if (tnOrig != null) {
                            // already was in typeNodes, remove
                            typeNodes.remove(tnOrig);
                        }

                        if ((pivot.costsFromSource + ts.getCosts()) < tn.costsFromSource) {
                            // should update costs - are now lower or were not
                            // set yet
                            tn.costsFromSource = pivot.costsFromSource
                                    + ts.getCosts();
                            tn.previous = pivot;
                            tn.transFromPrevious = ts;

                            //System.out.println(" ... updating
                            // costs="+tn.costsFromSource);//+"
                            // transFromPrevious="+tn.transFromPrevious+"
                            // previous="+tn.previous);
                        }

                        //System.out.print(" tn="+tn+" ts="+ts);
                        typeNodes.add(tn);
                        types2TypeNodes.put(tg, tn);
                    }

                    //System.out.println();
                }
            }

            //System.out.println("... examined, typeNodes# "+typeNodes.size());
            if (typeNodes.isEmpty()) {
                return sequence;
            }

            // pivot has the lowest costs, thus is first in the SortedSet typs
            pivot = (TypeNode) typeNodes.first();
            pivot.wasPivot = true;

            typeNodes.remove(pivot);

            //types2TypeNodes.remove(pivot.typ);
            //System.out.println("&&&&& at while end pivot.typ="+pivot.typ+"
            // target="+target);
        } while (!pivot.typ.equals(target));

        if (pivot.typ.equals(target)) {
            // FIXME
            //System.out.println("\n&&& found transf to type "+target+"
            // costs="+pivot.costsFromSource+"...");
            TypeNode p = pivot;

            while ((p != null) && (p.previous != null)) {
                sequence.add(0, p.transFromPrevious);

                //System.out.print(p.typ);
                p = p.previous;

                /*
                 * if(p == null) System.out.println("\n"); else
                 * System.out.print(" <- ");
                 */
            }
        }

        return sequence;
    }
}

class TypeNode implements Comparable {
    /** DOCUMENT ME! */
    protected TypeNode previous;

    /** DOCUMENT ME! */
    protected Type typ;

    /** DOCUMENT ME! */
    protected TypeTransformation transFromPrevious;

    /** DOCUMENT ME! */
    protected double costsFromSource;

    /** DOCUMENT ME! */
    protected boolean wasPivot;

    /**
     * Creates a new TypeNode object.
     * 
     * @param t DOCUMENT ME!
     */
    public TypeNode(Type t) {
        typ = t;
        transFromPrevious = null;
        costsFromSource = java.lang.Double.POSITIVE_INFINITY;
        previous = null;
        wasPivot = false;
    }

    /**
     * Creates a new TypeNode object.
     * 
     * @param t DOCUMENT ME!
     * @param ts DOCUMENT ME!
     */
    public TypeNode(Type t, TypeTransformation ts) {
        this(t);
        transFromPrevious = ts;
    }

    /**
     * DOCUMENT ME!
     * 
     * @param o DOCUMENT ME!
     * @return DOCUMENT ME!
     */
    public boolean equals(Object o) {
        TypeNode tn = (TypeNode) o;

        return typ.equals(tn.typ);
    }

    /**
     * DOCUMENT ME!
     * 
     * @return DOCUMENT ME!
     */
    public int hashCode() {
        return typ.hashCode();
    }

    /**
     * DOCUMENT ME!
     * 
     * @param o DOCUMENT ME!
     * @return DOCUMENT ME!
     */
    public int compareTo(Object o) {
        TypeNode te = (TypeNode) o;

        return ((int) (costsFromSource - te.costsFromSource) * 10)
                + (equals(o) ? 0 : 1);
    }

    /**
     * DOCUMENT ME!
     * 
     * @return DOCUMENT ME!
     */
    public String toString() {
        return "TypeNode[typ=" + typ + " previousTyp="
                + ((previous == null) ? null : previous.typ)
                + " transFromPrevious=" + transFromPrevious
                + " costsFromSource=" + costsFromSource;
    }
}

/*
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at http://www.mozilla.org/MPL/
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License. The
 * Original Code is: all this file. The Initial Developer of the Original Code
 * is: Tomas Pitner, Masaryk University in Brno, Czech Republic. Contributor(s):
 */
